from py_silhouette import SUPPORTED_DEVICE_PARAMETERS, DeviceState

class DummyDevice(object):
    """
    A dummy :py:class:`py_silhouette.Device` which, rather than driving an
    actual device, generates an SVG showing the cuts/lines plotted.
    """
    
    def __init__(self, filename=None, params=SUPPORTED_DEVICE_PARAMETERS[0]):
        """
        Parameters
        ----------
        filename : str or None
            If not None, the file to write an SVG to when :py:meth:`flush` is
            called.
        device_params : :py:class:`DeviceParameters`
            Definition of the device's key parameters.
        """
        
        self.filename = filename
        self.params = params
        
        self.width = params.area_width_max
        self.height = params.area_height_max
        
        self.regmarks_used = False
        self.speed = None
        self.force = None
        self.tool_diameter = None
        self.paths = [[(0, 0)]]  # [[(x, y), ...], ...]
    
    def zero_on_registration_mark(self, width, height, *args, **kwargs):
        self.width = width
        self.height = height
        
        self.regmarks_used = True
    
    def move_to(self, x, y, use_tool=False):
        if use_tool:
            self.paths[-1].append((x, y))
        else:
            self.paths.append([(x, y)])
    
    def move_home(self):
        self.move_to(0, 0, False)
    
    def set_speed(self, speed):
        self.speed = speed
    
    def set_force(self, force):
        self.force = force
    
    def set_tool_diameter(self, tool_diameter):
        self.tool_diameter = tool_diameter
    
    def get_state(self):
        return DeviceState.ready
    
    def flush(self):
        if self.filename is not None:
            with open(self.filename, "w") as f:
                f.write('<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n')
                f.write('<svg\n')
                f.write('  xmlns="http://www.w3.org/2000/svg"\n')
                f.write('  xmlns:inkscape="http://www.inkscape.org/namespaces/inkscape"\n')
                f.write('  width="{}mm"\n'.format(self.width))
                f.write('  height="{}mm"\n'.format(self.height))
                f.write('  viewBox="0 0 {} {}"\n'.format(self.width, self.height))
                f.write('>\n')
                
                f.write('  <!-- tool_diameter = {} mm -->\n'.format(
                    self.tool_diameter,
                ))
                f.write('  <!-- speed = {} mm/s -->\n'.format(
                    self.speed,
                ))
                f.write('  <!-- force = {} g -->\n'.format(
                    self.force,
                ))
                f.write('  <!-- regmarks_used = {} -->\n'.format(
                    self.regmarks_used,
                ))
                
                # Filter out non-drawn parts of the path
                paths = [path for path in self.paths if len(path) > 1]
                
                num_points = sum(len(p) - 1 for p in paths)
                i = 0
                for path in paths:
                    for (x1, y1), (x2, y2) in zip(path, path[1:]):
                        f.write('  <path\n')
                        f.write('    d="M{},{}L{},{}"\n'.format(
                            x1, y1,
                            x2, y2,
                        ))
                        f.write('    stroke="hsl({}, 100%, 50%)"\n'.format(
                            (360*i) // num_points
                        ))
                        f.write('    stroke-width="0.2"\n')
                        f.write('  />\n')
                        i += 1
                
                f.write('</svg>\n')



"""
This module includes a simple hueristic for guessing whether an
Inkscape-originated SVG should be plotted or cut.
"""

import re

from enum import Enum

from plottie.svg_utils import (
    is_inkscape_layer,
    get_inkscape_layer_label,
)


CUT_LAYER_NAMES_REGEX = re.compile(r"\bcut(s|ting|out)?|\bedges?",
                                   re.IGNORECASE)
"""
Regex matching layer names typically associated with cutting.
"""

PLOT_LAYER_NAMES_REGEX = re.compile(r"\bplo?ts?",
                                    re.IGNORECASE)
"""
Regex matching layer names typically associated with plotting.
"""

class PlotMode(Enum):
    """The plotting/cutting mode selected."""
    
    plot = "plot"
    cut = "cut"


def guess_plot_mode(svg):
    """
    Given an SVG as an ElementTree, guess the PlotMode to use based on the
    layer names preesnt. This is based on the names of Inkscape layers in the
    SVG.
    """
    layer_names = [
        get_inkscape_layer_label(node)
        for node in svg.iter()
        if is_inkscape_layer(node)
    ]
    
    maybe_cut = any(CUT_LAYER_NAMES_REGEX.search(name) is not None
                    for name in layer_names)
    maybe_plot = any(PLOT_LAYER_NAMES_REGEX.search(name) is not None
                     for name in layer_names)
    
    # Only guess if not ambiguous
    if maybe_cut and not maybe_plot:
        return PlotMode.cut
    elif maybe_plot and not maybe_cut:
        return PlotMode.plot
    else:
        return None



"""
Fairly generic XML utility functions.
"""

from xml.etree import ElementTree


def read_xml_file(filename, ignored_namespaces=[]):
    """
    Parse the provided XML file and return a
    :py:class:`xml.etree.ElementTree.Element`.
    
    For convenience, strips namespace prefixes from tag names  listed in the
    ``ignored_namespaces`` argument.
    """
    bracketed_namespaces = ["{{{}}}".format(ns) for ns in ignored_namespaces]
    
    with open(filename) as f:
        parser = ElementTree.iterparse(f)
        for _, element in parser:
            for namespace in bracketed_namespaces:
                if element.tag.startswith(namespace):
                    element.tag = element.tag[len(namespace):]
    
    return parser.root


def xml_deep_child_index(root, target):
    """
    Given an XML ElementTree and a child within that tree, return an 'index'
    into the tree.
    
    Indices are of the form [] indicates the root, [1] indicates the second
    child of the root and [1,2] indicates the third child of the second child
    of the root.
    
    Raises a KeyError if the target is not in the tree.
    """
    # Is root?
    if target is root:
        return []
    
    # Is direct child?
    for index, child in enumerate(root):
        if target is child:
            return [index]
    
    # Is descendent?
    for index, child in enumerate(root):
        try:
            return [index] + xml_deep_child_index(child, target)
        except KeyError:
            pass
    
    # Not found
    raise KeyError("{} not in {}".format(target, root))


def xml_get_at_index(root, index):
    """
    Given an index as produced by xml_deep_child_index, return the element at
    that index.
    """
    target = root
    for i in index:
        target = target[i]
    return target



"""
Functions for re-ordering/combining lines to reduce cutting and plotting time.
"""

from collections import defaultdict


def optimise_lines(lines, start_pos=(0, 0)):
    """
    Given a list of lines, return a new list of lines in an order which
    attempts to reduce the time spent moving between line endpoints or lifting
    the pen/cutter. Lines with coincident endpoints will be merged into a
    single contiguous line. 
    """
    
    # Enumerates the lines which may start or end at a given point.
    # {(x, y): [line_index, ...], ...}
    endpoints_to_lines = defaultdict(list)
    for line_index, line in enumerate(lines):
        if line[0] != line[-1]:
            # Open path (can only start/end at ends)
            endpoints_to_lines[line[0]].append(line_index)
            endpoints_to_lines[line[-1]].append(line_index)
        else:
            # Closed path (can start/end at any point)
            for point in line:
                endpoints_to_lines[point].append(line_index)
    
    # The optimised set of lines
    # [[(x, y), ...], ...]
    out = []
    
    cur_pos = start_pos
    for _ in range(len(lines)):
        # A simple greedly agorithm: pick the point nearest to our current
        # position and move to that
        endpoint = min(
            endpoints_to_lines,
            key=lambda ep: (ep[0] - cur_pos[0])**2 + (ep[1] - cur_pos[1])**2,
        )
        
        # Pick an arbitary line at this point
        next_line_index = endpoints_to_lines[endpoint][0]
        next_line = lines[next_line_index]
        
        if next_line[0] != next_line[-1]:
            # Open line segment.
            
            # Reverse line to connect to previous position more easily if necessary
            if next_line[0] == endpoint:
                next_line = next_line[:]
            else:
                next_line = next_line[::-1]
            
            # Remove this line from the lookup
            for point in [next_line[0], next_line[-1]]:
                endpoints_to_lines[point].remove(next_line_index)
                if not endpoints_to_lines[point]:
                    del endpoints_to_lines[point]
        else:
            # Closed line segment.
            
            # Remove the line from the lookup
            for point in next_line:
                endpoints_to_lines[point].remove(next_line_index)
                if not endpoints_to_lines[point]:
                    del endpoints_to_lines[point]
            
            # Re-order such that we start from the chosen point
            for start_index, point in enumerate(next_line):
                if point == endpoint:
                    break
            next_line = next_line[start_index:-1] + next_line[:start_index+1]
        
        # Add line to output (combining with previous line, if possible)
        if out and out[-1][-1] == endpoint:
            out[-1].extend(next_line[1:])
        else:
            out.append(next_line)
        
        cur_pos = next_line[-1]
    
    return out



"""
Group together line segments such that inner-most shapes are drawn first.
"""

from toposort import toposort


def find_dependencies(lines):
    """
    For each line in the input, return the indices of all lines which are
    completely enclosed by this line.
    
    Parameters
    ==========
    lines : [[(x, y), ...], ...]
    
    Returns
    =======
    dependencies : {line_index: {contained_line_index, ...}, ...}
    """
    bounding_boxes = [
        (
            min(x for x, y in line),
            min(y for x, y in line),
            max(x for x, y in line),
            max(y for x, y in line),
        )
        for line in lines
    ]
    
    return {
        outer_id: {
            inner_id
            for inner_id, (ix1, iy1, ix2, iy2) in enumerate(bounding_boxes)
            if (
                ox1 < ix1 and oy1 < iy1 and
                ox2 > ix2 and oy2 > iy2
            )
        }
        for outer_id, (ox1, oy1, ox2, oy2) in enumerate(bounding_boxes)
    }


def group_inside_first(lines):
    """
    Group the input lines into sets of lines which may be re-ordered without
    causing outer-most shapes to be drawn before inner-most shapes.
    """
    dependencies = find_dependencies(lines)
    
    out = []
    for line_indices in toposort(dependencies):
        group = []
        for index in sorted(line_indices):
            group.append(lines[index])
        out.append(group)
    
    return out



"""
Utility functions for finding registration marks in SVGs.

For a full description of the registration marks identified by this module see
the documentation for
:py:meth:`py_silhouette.SilhouetteDevice.zero_on_registration_mark`.
"""

import math

from attr import attrs, attrib


@attrs(frozen=True)
class RegmarkBox(object):
    """Parameters describing a regmark's top-left filled square box."""
    
    x = attrib()
    """
    X coordinate of the top-left corner of the mark (accounting for stroke
    width).
    """
    
    y = attrib()
    """
    Y coordinate of the top-left corner of the mark (accounting for stroke
    width).
    """
    
    size = attrib()
    """Width and height of the box (including stroke)."""
    
    thickness = attrib()
    """Stroke thickness of the box outline."""


@attrs(frozen=True)
class RegmarkBracket(object):
    """
    Base class for parameters describing a regmark's bottom-left or top-right
    bracket.
    """
    
    x = attrib()
    """
    X coordinate of the corner of the bracket (accounting for stroke width).
    """
    
    y = attrib()
    """
    Y coordinate of the corner of the bracket (accounting for stroke width).
    """
    
    length = attrib()
    """Length of the bracket's 'arms'."""
    
    thickness = attrib()
    """Stroke thickness of the bracket."""


class RegmarkBottomLeftBracket(RegmarkBracket):
    """Bottom left bracket dimensions."""

class RegmarkTopRightBracket(RegmarkBracket):
    """Bottom left bracket dimensions."""


def classify_regmark_component(colour, thickness, line):
    """
    For internal use. Given a line, classify what part of a registration mark
    it could be.
    
    .. note::
        This function will accept small floating point errors in line length
        euqality checks. The reported size/length of the returned feature is
        chosen arbitrarily from the measured values encountered.
        
        The function does *not* tolerate floating point errors in coordinates
        which should be coincident (i.e. the closing of a box) and horizontal
        and vertical lines *must* be exactly horizontal and vertical.
    
    Parameters
    ----------
    colour: (r, g, b, a) or None
    thickness: float or None
    line: [(x, y), ...]
    
    Returns
    -------
    regmark_type
        One of:
        
        * `RegmarkBox`
        * `RegmarkBottomLeftBracket`
        * `RegmarkTopRightBracket`
        * `None`
    """
    # Only consider lines with a simple, black stroke
    if colour is None or colour != (0, 0, 0, 1):
        return None
    if thickness is None or thickness == 0:
        return None
    
    # Don't bother considering at anything other than closed squares and
    # brackets described with the minimum number of lines
    if len(line) != 3 and len(line) != 5:
        return None
    
    # Check that the orientation of lines alternates between horizontal and
    # vertical and that all lines are the same length
    last_is_horizontal = None
    last_length = None
    segment_polarities = []  # 'True' if +ve, 'False' if -ve
    for (x1, y1), (x2, y2) in zip(line[0:], line[1:]):
        is_horizontal = y1 == y2
        is_vertical = x1 == x2
        
        # We've found a non-horizontal/vertical line which definately isn't
        # part of a registration mark.
        if is_horizontal is is_vertical:
            return None
        
        # Make sure the line orientation alternates
        if last_is_horizontal is None:
            last_is_horizontal = is_horizontal
        elif (last_is_horizontal == is_horizontal or
                last_is_horizontal != is_vertical):
            # Repeated horizontal/vertical line: not a regmark.
            return None
        last_is_horizontal = is_horizontal
        
        # Make sure the lines are the same length
        length = x2 - x1 if is_horizontal else y2 - y1
        if last_length is None:
            last_length = abs(length)
        elif not math.isclose(last_length, abs(length), rel_tol=1e-6):
            return None
        
        segment_polarities.append(length >= 0)
    
    if len(line) == 3:
        # This is a bracket, determine which type
        #
        # Valid brackets:
        #
        #                                           222      111
        #                         1        2           1        2
        #                         1        2           1        2
        #                          222      111
        #
        #     last_is_horizontal  T        F        T        F
        #     segment_polarities  TT       FF       FF       TT
        #
        # Invalid brackets:
        #
        #                                            222      111
        #                             1        2    1        2
        #                             1        2    1        2
        #                          222      111
        #
        #     last_is_horizontal  T        F        T        F
        #     segment_polarities  TF       TF       FT       FT
        p1, p2 = segment_polarities
        if last_is_horizontal == p1 == p2:
            return RegmarkBottomLeftBracket(
                x=min(line[0][0], line[2][0]) - (thickness/2.0),
                y=max(line[0][1], line[2][1]) + (thickness/2.0),
                length=last_length,
                thickness=thickness,
            )
        elif (not last_is_horizontal) == p1 == p2:
            return RegmarkTopRightBracket(
                x=max(line[0][0], line[2][0]) + (thickness/2.0),
                y=min(line[0][1], line[2][1]) - (thickness/2.0),
                length=last_length,
                thickness=thickness,
            )
        else:
            return None
    elif len(line) == 5:
        # If the line is closed, this *must* be a box since all sides are
        # perpendicular.
        if line[0] == line[4]:
            return RegmarkBox(
                x=min(line[0][0], line[2][0]) - (thickness/2.0),
                y=min(line[0][1], line[2][1]) - (thickness/2.0),
                size=last_length + thickness,
                thickness=thickness,
            )
        else:
            return None
        


@attrs
class RegmarkSpecification(object):
    """
    A specification of the key measurements of a set of registration marks.
    """
    
    x = attrib()
    """Global X coordinate (mm) of the top-left corner of the box."""
    
    y = attrib()
    """Global Y coordinate (mm) of the top-left corner of the box."""
    
    width = attrib()
    """Width (mm) of the registered area (to outermost bounds of the mark)."""
    
    height = attrib()
    """Height (mm) of the registered area (to outermost bounds of the mark)."""
    
    box_size = attrib(default=5.0)
    """The size of the square box (mm), including strokes."""
    
    line_length = attrib(default=20.0)
    """The length of the bracket lines (mm)."""
    
    line_thickness = attrib(default=0.5)
    """The thickness of the bracket line strokes (mm)."""
    
    def is_line_part_of_regmark(self, colour, thickness, line):
        """
        Is the specified line part of this registration mark?
        
        .. note:

            This function will tolerate small errors of 1e-6 or less in line
            lengths to handle floating point errors in the input.
        
        .. note:

            If a document (eroneously) has several overlapping registration
            marks with differing box-sizes/lengths/thicknesses brackets which
            are entirely covered by the brackets in this regmark are also
            considered part of the regmark.
        
        Parameters
        ----------
        colour: (r, g, b, a) or None
        thickness: float or None
        line: [(x, y), ...]
        
        Returns
        -------
        is_part_of_regmark : bool
        """
        c = classify_regmark_component(colour, thickness, line)
        
        if c is None:
            return False
        elif isinstance(c, RegmarkBox):
            return (
                math.isclose(c.x, self.x, rel_tol=1e-6) and
                math.isclose(c.y, self.y, rel_tol=1e-6) and
                c.size <= (self.box_size * (1.0+1e-6))
            )
        elif isinstance(c, RegmarkBottomLeftBracket):
            return (
                math.isclose(c.x, self.x, rel_tol=1e-6) and
                math.isclose(c.y, self.y + self.height, rel_tol=1e-6) and
                c.length <= (self.line_length * (1.0+1e-6)) and
                c.thickness <= (self.line_thickness * (1.0+1e-6))
            )
        elif isinstance(c, RegmarkTopRightBracket):
            return (
                math.isclose(c.x, self.x + self.width, rel_tol=1e-6) and
                math.isclose(c.y, self.y, rel_tol=1e-6) and
                c.length <= (self.line_length * (1.0+1e-6)) and
                c.thickness <= (self.line_thickness * (1.0+1e-6))
            )
        else:
            assert False


def find_regmarks(outlines,
                  required_box_size=None,
                  required_line_length=None,
                  required_line_thickness=None):
    """
    Find a set of registration marks given a list of outlines from an SVG (e.g.
    extracted by :py:func:`svgoutline.svg_to_outlines`).
    
    .. note::
    
        Because this function only takes outlines as an argument, the
        upper-left box in the registration mark *must* be stroked for it to be
        found.
    
    Parameters
    ----------
    outlines : [(colour, thickness, line), ...]
        A series of lines from an SVG. ``colour`` should be a tuple (r, g, b,
        a) or None. ``thickness`` should be a float or None. ``line`` should be
        a list of (x, y) tuples.
        
        The provided registration mark lines are allowed to be misaligend by a
        factor of 1e-6 to tolerate floating point errors during their
        transformation. Horizontal and vertical lines and coincident points
        *must* be numerically exactly horizontal, vertical and coincident
        respectively.
    required_box_size : float or None
        If specified, ignore registration marks whose upper-left box is not
        this size. This quantity is allowed to differ from the requred value by
        a relative factor of 1e-6 to tolerate floating point errors in
        the generation of the outline list.
    required_line_length : float or None
        If specified, ignore registration marks whose brackets are not this
        length. This quantity is allowed to differ from the requred value by
        a relative factor of 1e-6 to tolerate floating point errors in
        the generation of the outline list.
    required_line_thickness : float or None
        If specified, ignore registration marks whose brackets are not this
        thickness. This quantity is allowed to differ from the requred value by
        a relative factor of 1e-6 to tolerate floating point errors in
        the generation of the outline list.

    Returns
    -------
    spec : :py:class:`RegmarkSpecification` or None
        If registration marks are found, a description of those marks is
        returned. If multiple sets of registration marks are found, the
        outermost set will be assumed.
        
        If no regmarks were found, None is returned.
    """
    # [RegmarkBox, ...]
    boxes = []
    
    # [RegmarkBottomLeftBracket, ...]
    bl_brackets = []
    
    # [RegmarkTopRightBracket, ...]
    tr_brackets = []
    
    for colour, thickness, line in outlines:
        c = classify_regmark_component(colour, thickness, line)
        if isinstance(c, RegmarkBox) and (
                required_box_size is None or
                math.isclose(c.size, required_box_size, rel_tol=1e-6)):
            boxes.append(c)
        elif isinstance(c, RegmarkBracket) and (
                (required_line_length is None or
                    math.isclose(c.length, required_line_length, rel_tol=1e-6)) and
                (required_line_thickness is None or
                    math.isclose(c.thickness, required_line_thickness, rel_tol=1e-6))):
            if isinstance(c, RegmarkBottomLeftBracket):
                bl_brackets.append(c)
            if isinstance(c, RegmarkTopRightBracket):
                tr_brackets.append(c)
    
    # Find all matching pairings of boxes and brackets
    #
    # [RegmarkSpecification, ...]
    regmarks = [
        RegmarkSpecification(
            x=box.x,
            y=box.y,
            width=tr_bracket.x - box.x,
            height=bl_bracket.y - box.y,
            box_size=box.size,
            line_length=bl_bracket.length,
            line_thickness=bl_bracket.thickness,
        )
        for box in boxes
        for bl_bracket in bl_brackets
        for tr_bracket in tr_brackets
        if (
            # Bottom left bracket should be at the bottom-left corner
            math.isclose(box.x, bl_bracket.x, rel_tol=1e-6) and
            box.y + box.size < bl_bracket.y - bl_bracket.length and
            # Top right bracket should be at the bottom-left corner
            math.isclose(box.y, tr_bracket.y, rel_tol=1e-6) and
            box.x + box.size < tr_bracket.x - bl_bracket.length and
            # Brackets should be same size/thickness
            math.isclose(bl_bracket.length, tr_bracket.length, rel_tol=1e-6) and
            math.isclose(bl_bracket.thickness, tr_bracket.thickness, rel_tol=1e-6)
        )
    ]
    
    # Return the outer-most regmarks found
    if not regmarks:
        None
    else:
        return min(regmarks, key=lambda r: (
            # Outermost
            r.x, r.y,
            -r.width, -r.height,
            # With longest, thickest lines
            -r.line_length,
            -r.line_thickness,
        ))



"""
Functions to over-cut closed line segments slightly to ensure a complete cut.
"""

import math


def first_part_of_line(line, distance):
    """
    Given a line defined by [(x, y), ...], return a line which follows the
    first 'distance' along the line.
    """
    # Special case
    if distance <= 0:
        return []

    out = []
    for ((x1, y1), (x2, y2)) in zip(line, line[1:]):
        out.append((x1, y1))
        
        dx = x2 - x1
        dy = y2 - y1
        length = math.sqrt((dx*dx) + (dy*dy))

        if length < distance:
            distance -= length
        else:
            xm = x1 + (dx * (distance / length))
            ym = y1 + (dy * (distance / length))
            out.append((xm, ym))
            break
    else:
        # Distance is > length of line, don't go around again, just leave it at
        # the end.
        out.extend(line[-1:])

    return out


def over_cut_lines(lines, distance=1.0):
    """
    Over-cut closed lines to ensure a complete cut.
    
    Parameters
    ==========
    lines : [[(x ,y), ...], ...]
        The lines to be cut.
    distance : float
        The distance to over-cut by.
    
    Returns
    =======
    new_lines : [[(x ,y), ...], ...]
    """
    out = []
    
    for line in lines:
        if line[0] != line[-1]:
            # Not a closed cut
            out.append(line)
        else:
            # Closed line, extend as required
            out.append(line + first_part_of_line(line, distance)[1:])
    
    return out



"""
Utilities for manipulating/querying aspects of SVGs relevant to this
application (not a generic SVG library). All methods expect SVGs parsed as XML
using Python's ElementTree library.
"""

import re

from xml.etree import ElementTree

from collections import OrderedDict

from colorsys import hls_to_rgb

from plottie.xml_utils import xml_deep_child_index, xml_get_at_index
from plottie.css_colour_names import CSS_COLOUR_NAMES


# Relevant XML namespace URIs used by SVGs
SVG_NAMESPACE = "http://www.w3.org/2000/svg"
INKSCAPE_NAMESPACE = "http://www.inkscape.org/namespaces/inkscape"
XLINK_NAMESPACE = "http://www.w3.org/1999/xlink"

namespaces = {
    "svg": SVG_NAMESPACE,
    "inkscape": INKSCAPE_NAMESPACE,
    "xlink": XLINK_NAMESPACE,
}

# Unit conversion ratios
MM_PER_CM = 10.0
MM_PER_QUARTER_MM = 0.25
MM_PER_INCH = 25.4
MM_PER_PICA = MM_PER_INCH / 6.0
MM_PER_POINT = MM_PER_INCH / 72.0

CSS_COLOUR_HEX3 = re.compile(r"^#"
                             r"([0-9a-f])"
                             r"([0-9a-f])"
                             r"([0-9a-f])"
                             r"$", re.I)
CSS_COLOUR_HEX4 = re.compile(r"^#"
                             r"([0-9a-f])"
                             r"([0-9a-f])"
                             r"([0-9a-f])"
                             r"([0-9a-f])"
                             r"$", re.I)
CSS_COLOUR_HEX6 = re.compile(r"^#"
                             r"([0-9a-f]{2})"
                             r"([0-9a-f]{2})"
                             r"([0-9a-f]{2})"
                             r"$", re.I)
CSS_COLOUR_HEX8 = re.compile(r"^#"
                             r"([0-9a-f]{2})"
                             r"([0-9a-f]{2})"
                             r"([0-9a-f]{2})"
                             r"([0-9a-f]{2})"
                             r"$", re.I)
CSS_COLOUR_RGB = re.compile(r"^rgb\s*[(]\s*"
                            r"([0-9]{1,3})\s*,\s*"
                            r"([0-9]{1,3})\s*,\s*"
                            r"([0-9]{1,3})\s*"
                            r"[)]$", re.I)
CSS_COLOUR_RGB_PERC = re.compile(r"^rgb\s*[(]\s*"
                                 r"([0-9]{1,3}(?:[.][0-9]*)?)\s*%\s*,\s*"
                                 r"([0-9]{1,3}(?:[.][0-9]*)?)\s*%\s*,\s*"
                                 r"([0-9]{1,3}(?:[.][0-9]*)?)\s*%\s*"
                                 r"[)]$", re.I)
CSS_COLOUR_RGBA = re.compile(r"^rgba\s*[(]\s*"
                             r"([0-9]{1,3})\s*,\s*"
                             r"([0-9]{1,3})\s*,\s*"
                             r"([0-9]{1,3})\s*,\s*"
                             r"([0-9]+|[0-9]+[.][0-9]*|[.][0-9]+)\s*"
                             r"[)]$", re.I)
CSS_COLOUR_RGBA_PERC = re.compile(r"^rgba\s*[(]\s*"
                                  r"([0-9]{1,3}(?:[.][0-9]*)?)\s*%\s*,\s*"
                                  r"([0-9]{1,3}(?:[.][0-9]*)?)\s*%\s*,\s*"
                                  r"([0-9]{1,3}(?:[.][0-9]*)?)\s*%\s*,\s*"
                                  r"([0-9]+|[0-9]+[.][0-9]*|[.][0-9]+)\s*"
                                  r"[)]$", re.I)
CSS_COLOUR_HSL_PERC = re.compile(r"^hsl\s*[(]\s*"
                                 r"([0-9]{1,3})\s*,\s*"
                                 r"([0-9]{1,3}(?:[.][0-9]*)?)\s*%\s*,\s*"
                                 r"([0-9]{1,3}(?:[.][0-9]*)?)\s*%\s*"
                                 r"[)]$", re.I)
CSS_COLOUR_HSLA_PERC = re.compile(r"^hsla\s*[(]\s*"
                                  r"([0-9]{1,3})\s*,\s*"
                                  r"([0-9]{1,3}(?:[.][0-9]*)?)\s*%\s*,\s*"
                                  r"([0-9]{1,3}(?:[.][0-9]*)?)\s*%\s*,\s*"
                                  r"([0-9]+|[0-9]+[.][0-9]*|[.][0-9]+)\s*"
                                  r"[)]$", re.I)


def css_dimension_to_mm(dimension, pixels_per_mm=96.0/MM_PER_INCH,
                        width_mm=None, height_mm=None):
    """
    Convert a CSS dimension string (e.g. '3cm') into a number of mm. Fails for
    viewport-precentage length units of size if 'width_mm' and 'height_mm' are
    None.
    """
    match = re.match(r"^\s*[+]?\s*([0-9.]+)\s*(vw|vh|vmin|vmax|cm|mm|Q|in|pc|pt|px|)\s*$", dimension)
    if not match:
        raise ValueError("{} is not a positive, absolute unit of size".format(
            repr(dimension)))
    
    number, unit = match.groups()
    number = float(number)
    
    if unit == "mm":
        pass
    elif unit == "cm":
        number *= MM_PER_CM
    elif unit == "Q":
        number *= MM_PER_QUARTER_MM
    elif unit == "in":
        number *= MM_PER_INCH
    elif unit == "pc":
        number *= MM_PER_PICA
    elif unit == "pt":
        number *= MM_PER_POINT
    elif unit == "px" or unit == "":
        number /= pixels_per_mm
    elif unit == "vw":
        if width_mm is None:
            raise ValueError("Relative dimensions not allowed.")
        number = (number / 100.0) * width_mm
    elif unit == "vh":
        if height_mm is None:
            raise ValueError("Relative dimensions not allowed.")
        number = (number / 100.0) * height_mm
    elif unit == "vmin":
        if width_mm is None or height_mm is None:
            raise ValueError("Relative dimensions not allowed.")
        number = (number / 100.0) * min(width_mm, height_mm)
    elif unit == "vmax":
        if width_mm is None or height_mm is None:
            raise ValueError("Relative dimensions not allowed.")
        number = (number / 100.0) * max(width_mm, height_mm)
    else:
        assert False, "Supposedly-supported unit not implemented!"
    
    return number


def css_colour_to_rgba(colour_string):
    """
    Parses colours according to the SVG/CSS specification (see
    https://www.w3.org/TR/css-color-3/#svg-color) into (r, g, b, a) tuples with
    all values being in the range 0.0 to 1.0. Also parses non-standard 4 and 8
    didgit hex colour codes with alpha information in the final nyble/byte (as
    used in Inkscape).
    """
    colour_string = colour_string.strip().lower()
    
    # If we've got a named colour, swap it for a concrete colour string
    colour_string = CSS_COLOUR_NAMES.get(colour_string, colour_string)
    
    # Normalise hex representations into 8-digit hex (with alpha)
    match = CSS_COLOUR_HEX3.match(colour_string)
    if match:
        r, g, b = match.groups()
        colour_string = "#{}{}{}ff".format(r*2, g*2, b*2)
    match = CSS_COLOUR_HEX4.match(colour_string)
    if match:
        r, g, b, a = match.groups()
        colour_string = "#{}{}{}{}".format(r*2, g*2, b*2, a*2)
    if CSS_COLOUR_HEX6.match(colour_string):
        colour_string = "{}ff".format(colour_string)
    
    # Parse hex colour codes
    match = CSS_COLOUR_HEX8.match(colour_string)
    if match:
        return tuple(int(x, 16)/255.0 for x in match.groups())
    
    # Normalise RGB into RGBA
    match = CSS_COLOUR_RGB.match(colour_string)
    if match:
        r, g, b = match.groups()
        colour_string = "rgba({},{},{},1)".format(r, g, b)
    match = CSS_COLOUR_RGB_PERC.match(colour_string)
    if match:
        r, g, b = match.groups()
        colour_string = "rgba({}%,{}%,{}%,1)".format(r, g, b)
    
    # Parse RGBA colour codes
    match = CSS_COLOUR_RGBA.match(colour_string)
    if match:
        r, g, b, a = match.groups()
        rgba = (int(r)/255.0, int(g)/255.0, int(b)/255.0, float(a))
        if max(rgba) > 1.0:
            raise ValueError("Colour value out of range.")
        return rgba
    match = CSS_COLOUR_RGBA_PERC.match(colour_string)
    if match:
        r, g, b, a = match.groups()
        rgba = (float(r)/100.0, float(g)/100.0, float(b)/100.0, float(a))
        if max(rgba) > 1.0:
            raise ValueError("Colour value out of range.")
        return rgba
    
    # Normalise HSL into HSLA
    match = CSS_COLOUR_HSL_PERC.match(colour_string)
    if match:
        h, s, l = match.groups()
        colour_string = "hsla({},{}%,{}%,1)".format(h, s, l)
    
    # Parse HSLA colour codes
    match = CSS_COLOUR_HSLA_PERC.match(colour_string)
    if match:
        h, s, l, a = map(float, match.groups())
        h /= 360.0
        s /= 100.0
        l /= 100.0
        if h >= 1.0 or max(s, l, a) > 1.0:
            raise ValueError("Colour value out of range.")
        r, g, b = hls_to_rgb(h, l, s)
        return (r, g, b, a)
    
    raise ValueError("Unrecognised colour format or name.")


def is_inkscape_layer(tag):
    """
    Is a given XML tag an Inkscape layer?
    """
    return (tag.tag in ("g", "{{{}}}g".format(SVG_NAMESPACE)) and
            tag.attrib.get("{{{}}}groupmode".format(INKSCAPE_NAMESPACE)) == "layer")


def get_inkscape_layer_label(tag):
    """
    Given an inkscape layer 'g' tag, return the label assigned to that layer.
    """
    return tag.attrib.get("{{{}}}label".format(INKSCAPE_NAMESPACE))


def is_visible(tag):
    """
    Given an svg tag, return True if it is visible and False if hidden.
    """
    style = tag.attrib.get("style")
    if style is None:
        return True
    else:
        match = re.match(r"(^|.*;)\s*display\s*:\s*none\s*($|;.*)", style)
        return match is None


def set_visibility(tag, visibility):
    """
    Given an SVG tag, mutate it to add or remove 'display:none' to the style
    attribute.
    """
    style = tag.attrib.get("style", "")
    
    # Remove visibility specifier
    style = re.sub(r"(^|;+)\s*display\s*:\s*[^\s;]+\s*($|;)", ";", style)
    style = style.strip("; \t\n")
    
    # Change/remove the style attribute
    if not visibility:
        style += ";display:none"
        style = style.strip(";")
    
    if style:
        tag.attrib["style"] = style
    else:
        tag.attrib.pop("style", None)


def make_nodes_visible(root, predicate):
    """
    Given an SVG document root and a predicate function, make all elements
    which the predicate returns True for visible and all others hidden.
    """
    if predicate(root):
        set_visibility(root, True)
        return True
    else:
        child_is_visible = False
        for child in root:
            child_is_visible |= make_nodes_visible(child, predicate)
        set_visibility(root, child_is_visible)
        return child_is_visible



from setuptools import setup, find_packages

with open("plottie/version.py", "r") as f:
    exec(f.read())

setup(
    name="plottie",
    version=__version__,
    packages=find_packages(),

    # Metadata for PyPi
    url="https://github.com/mossblaser/plottie",
    author="Jonathan Heathcote",
    description="Command-line software for plotting and cutting using Silhouette devices.",
    license="LGPLv3",
    classifiers=[
        "Development Status :: 3 - Alpha",

        "Intended Audience :: Developers",

        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",

        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",

        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
    ],
    keywords="svg plotter cutter silhouette",

    install_requires=["svgoutline", "py_silhouette", "attrs", "toposort", "enum34"],
    
    entry_points={
        "console_scripts": [
            "plottie = plottie.cli:main",
        ],
    }
)



"""
The ``plottie`` command line interface.
"""

import sys

import re

import time

from xml.etree import ElementTree

from argparse import ArgumentParser

from functools import partial

from svgoutline import svg_to_outlines, get_svg_page_size

import py_silhouette

from plottie.dummy_device import DummyDevice

from plottie.plot_mode_heuristics import (
    CUT_LAYER_NAMES_REGEX,
    PLOT_LAYER_NAMES_REGEX,
    PlotMode,
    guess_plot_mode,
)

from plottie.svg_utils import (
    is_inkscape_layer,
    get_inkscape_layer_label,
    css_dimension_to_mm,
    css_colour_to_rgba,
    make_nodes_visible,
)

from plottie.regmarks import (
    RegmarkSpecification,
    find_regmarks,
)

from plottie.inside_first import group_inside_first

from plottie.line_ordering import optimise_lines

from plottie.over_cut import over_cut_lines

import logging


DEFAULT_TOOL_FORCE_PERC = 0.2
"""
Default force to use (in range 0 to 1, a proportion of the device's supported
range).
"""

DEFAULT_TOOL_SPEED_PERC = 1.0
"""
Default speed to use (in range 0 to 1, a proportion of the device's supported
range).
"""

DEFAULT_OVER_CUT = 1.0
"""
Default distance to over-cut closed shapes in cut mode (mm).
"""


def make_layer_matcher(regex):
    """
    Return a function which takes a ElementTree node and tests whether it is an
    Inkscape layer with a name matching the specified regex.
    """
    regex = re.compile(regex, re.IGNORECASE)
    
    def matcher(node):
        if not is_inkscape_layer(node):
            return False
        if not regex.match(get_inkscape_layer_label(node)):
            return False
        return True
    
    return matcher


def make_id_matcher(id):
    """
    Return a function which takes a ElementTree node and tests whether it has
    the specified ID.
    """
    def matcher(node):
        return node.attrib.get("id") == id
    
    return matcher


def make_class_matcher(classname):
    """
    Return a function which takes a ElementTree node and tests whether it has
    the specified class.
    """
    def matcher(node):
        return classname in node.attrib.get("class", "").split()
    
    return matcher


def make_argument_parser():
    """
    Return a fully configured :py:class:`argparse.ArgumentParser` instance.
    """
    parser = ArgumentParser(description="""
        Plot and cut SVG files using a Silhouette desktop plotter/cutter.
    """)
    
    parser.add_argument(
        "svg", nargs="?",
        help="""
            The SVG file to plot or cut.
        """,
    )
    action = parser.add_argument_group(
        "action arguments",
        description="""
            Select what mode the plotter should use: cutting or plotting mode.
            If neither argument is used, plottie will attempt to guess whether
            cutting or plotting is required based on Inkscape layer names found
            in the input SVG and, failing that, will assume cutting.  In
            cutting mode, additional movements will be made by the plotter to
            compensate for the movement of the knife within the cartridge.
        """
    ).add_mutually_exclusive_group()
    
    action.add_argument(
        "--cut", "-c",
        action="store_const", dest="plot_mode", const=PlotMode.cut,
    )
    
    action.add_argument(
        "--plot", "-p",
        action="store_const", dest="plot_mode", const=PlotMode.plot,
    )
    
    # #########################################################################
    
    device_selection = parser.add_argument_group(
        "device selection arguments",
    )
    
    device_selection.add_argument(
        "--device", "-d",
        type=str, default="0",
        help="""
            Specify the name (or number) of the device to use. (See
            --list-devices for device names and numbers). If this argument is
            not given, the first device discovered will be used.
        """
    )
    
    device_selection.add_argument(
        "--list-devices", "--ls",
        action="store_true",
        help="""
            Discover and list the devices connected to this machine and exit.
        """
    )
    
    device_selection.add_argument(
        "--use-dummy-device", "-D",
        nargs="?", metavar="FILENAME", default=False,
        help="""
            If this option is given, a dummy device (called 'Dummy Device')
            will be added to the list of discovered devices and will be used by
            default. When a filename is given, the cutting or plotting commands
            sent to the device will be written to the specified file in SVG
            format. Otherwise, all cutting and plotting commands will be
            ignored.
        """
    )
    
    # #########################################################################
    
    object_visibility = parser.add_argument_group(
        "object visibility arguments",
        description="""
            These options may be used to control which parts of the SVG will be
            plotted or cut out. The default behaviour for Inkscape SVGs is to
            select only layers whose names match the regular expression '{}'
            when cutting or '{}' when plotting (as if specified via --layer).
            If no matching Inkscape layers are found, `--all` mode is used.
            Alternatively, the parts of the SVG to plot may be specified
            explicitly using a combination of the following arguments.
        """.format(
            CUT_LAYER_NAMES_REGEX.pattern,
            PLOT_LAYER_NAMES_REGEX.pattern,
        )
    )
    
    object_visibility.add_argument(
        "--all", "-a",
        action="store_true",
        help="""
            Plot or cut outlines visible in the unchanged SVG.
        """,
    )
    
    object_visibility.add_argument(
        "--layer", "-l",
        type=make_layer_matcher, metavar="REGEX",
        dest="visible_object_matchers", action="append",
        help="""
            Plot or cut visible outlines only on the Inkscape layers whose
            names match the provided case-insensitive regex. If any matching
            layers are not visible, they will be made visible. May be given
            several times to select multiple layers to plot.
        """,
    )
    
    object_visibility.add_argument(
        "--id", "-i",
        type=make_id_matcher,
        dest="visible_object_matchers", action="append",
        help="""
            Plot or cut only visible outlines for SVG elements which have the
            specified object ID (or are visible children of that object). May
            be given several times to select multiple IDs to plot.
        """,
    )
    
    object_visibility.add_argument(
        "--class", "-s",
        type=make_class_matcher,
        dest="visible_object_matchers", action="append",
        help="""
            Plot or cut only visible outlines for SVG elements which have the
            specified class (or are visible children of an object with that
            class). May be given several times to select multiple IDs to plot.
        """,
    )
    
    # #########################################################################
    
    line_selection = parser.add_argument_group("line selection arguments")
    
    line_selection.add_argument(
        "--colour", "--color", "-C",
        type=css_colour_to_rgba, action="append",
        help="""
            Plot or cut only outlines with the listed stroke colour. Accepts
            CSS-style colour codes. May be given multiple times to include
            lines with multiple stroke colours. Lines with gradient or
            patterned strokes will be ignored. If not specified, all stroked
            lines (of all colours) will be plotted or cut.
        """,
    )
    
    # #########################################################################
    
    plotting_parameters = parser.add_argument_group(
        "plotting parameter arguments",
    )
    
    plotting_parameters.add_argument(
        "--speed", "-S",
        type=str, default="{:.0f}%".format(DEFAULT_TOOL_SPEED_PERC*100),
        help="""
            The speed with which to plot or cut. Either a number
            given in mm/s or a percentage ending with a '%%' symbol giving the
            percentage of the supported speed to run at. Defaults to
            %(default)s.
        """,
    )
    
    plotting_parameters.add_argument(
        "--force", "-F",
        type=str, default="{:.0f}%".format(DEFAULT_TOOL_FORCE_PERC*100),
        help="""
            The force with which to plot or cut. Either a number given in grams
            or a percentage ending with a '%%' symbol giving the percentage of
            the maximum supported force to apply. Defaults to
            %(default)s.
        """,
    )
    
    # #########################################################################
    
    ordering_parameters = parser.add_argument_group("plotting and cutting order arguments")
    
    ordering_parameters.add_argument(
        "--inside-first",
        action="store_true", dest="inside_first", default=None,
        help="""
            If some lines and shapes are contained within others, cut or plot
            these first. Inside first mode will be disabled by default in
            `--plot` mode and enabled by default for `--cut` mode. See also
            `--no-inside-first`.
        """,
    )
    
    ordering_parameters.add_argument(
        "--no-inside-first",
        action="store_false", dest="inside_first",
        help="""
            Don't try to plot or cut lines and shapes contained within others
            first. See also `--inside-first`.
        """,
    )
    
    ordering_parameters.add_argument(
        "--fast-order",
        action="store_true", dest="fast_order", default=True,
        help="""
            If specified, plot or cut paths in an order which attempts to
            reduce the time spent moving. Fast ordering is used by default.
            Opposite of `--native-order`.
        """,
    )
    
    ordering_parameters.add_argument(
        "--native-order",
        action="store_false", dest="fast_order",
        help="""
            If specified, plot/cuts paths in the order they appear in the input
            file.  Opposite of `--fast-order`.
        """,
    )
    
    ordering_parameters.add_argument(
        "--over-cut",
        nargs="?", type=float, const=DEFAULT_OVER_CUT, metavar="DISTANCE",
        dest="over_cut", default=None,
        help="""
            Extend closed shapes such that the first 'DISTANCE' mm of the shape
            is plotted/cut twice. If no value is given, %(const)0.1f mm is
            assumed. This option is enabled by default in --cut mode and
            disabled by default in --plot mode.
        """,
    )
    
    ordering_parameters.add_argument(
        "--no-over-cut",
        action="store_const", const=0.0, dest="over_cut",
        help="""
            Disable --over-cut mode.
        """,
    )
    
    # #########################################################################
    
    registration_parameters = parser.add_argument_group("registration arguments")
    
    registration_parameters.add_argument(
        "--regmarks", "-r",
        action="store_true", dest="regmarks", default=None,
        help="""
            Force detection of registration marks in the input SVG and throw an
            error if none are found. (By default registration marks will be
            used if found but no error is thrown if they are not). The
            registration marks must be visible and drawn as a 5x5mm
            stroked-and-filled black square (accounting for stroke thickness)
            at the top left and a pair of 'L' brackets 20x20mm at the bottom
            left and top right. Strokes should be black and 0.5mm thick. The
            registration marks will be excluded from plotting or cutting
            commands unless `--include-regmarks` is used.  This is the default
            mode if registration marks are detected in the design.
        """,
    )
    
    registration_parameters.add_argument(
        "--no-regmarks", "-R",
        action="store_false", dest="regmarks",
        help="""
            Disables auto-detection of registration marks in the input SVG.
        """,
    )
    
    registration_parameters.add_argument(
        "--manual-regmarks",
        nargs=4,
        metavar=("X-OFFSET", "Y-OFFSET", "WIDTH", "HEIGHT"),
        dest="regmarks",
        help="""
            Command the plotter to find registration marks denoting an area WIDTH x
            HEIGHT which is (X-OFFSET, Y-OFFSET) from the top-left of the input
            file. CSS-style absolute units should be used for all coordinates and
            dimensions (e.g.  '3mm'). Registration marks at the specified
            coordinates will be automatically removed from plotting/cutting paths
            unless `--include-regmarks` is used. The registration marks will be
            assumed to have 20mm brackets with 0.5mm black strokes.
        """,
    )
    
    registration_parameters.add_argument(
        "--include-regmarks",
        action="store_true", dest="include_regmarks", default=False,
        help="""
            Don't remove registration marks from the SVG before plotting or
            cutting.
        """,
    )
    
    return parser


def enumerate_devices(include_dummy=False, dummy_filename=None):
    """
    Enumerate all connected devices (including a dummy device, if specified).
    
    Returns
    =======
    [(name, func), ...]
        A list of names and :py:class:`py_silhouette.Device`-creating
        functions.
    """
    devices = []
    
    for device, params in py_silhouette.enumerate_devices():
        devices.append((
            params.product_name,
            partial(py_silhouette.SilhouetteDevice, device, params),
        ))
    
    if include_dummy:
        devices.insert(0, (
            "Dummy Device",
            partial(DummyDevice, dummy_filename),
        ))
    
    return devices


def print_device_list(include_dummy=False):
    """
    Print a (numbered) list of connected devices to stdout.
    """
    for number, (name, _) in enumerate(enumerate_devices(include_dummy)):
        print("{}: {}".format(number, name))


def parse_svg_argument(parser, args):
    if args.svg is None:
        parser.error("SVG filename required.")
    try:
        args.svg = ElementTree.parse(args.svg).getroot()
    except (IOError, SystemError):
        parser.error("{} does not exist or cannot be read".format(args.svg))
    except ElementTree.ParseError as e:
        parser.error("SVG file must be valid XML: {}".format(str(e)))
    logging.info("Loaded SVG")


def parse_device_arguments(parser, args):
    devices = enumerate_devices(
        args.use_dummy_device is not False,
        args.use_dummy_device,
    )
    if len(devices) == 0:
        parser.error("No connected devices found.")
    
    device_name = None
    connect_fn = None
    try:
        device_number = int(args.device)
        try:
            device_name, connect_fn = devices[device_number]
        except IndexError:
            pass
    except ValueError:
        device_string = args.device
        for name, fn in devices:
            if name.lower().startswith(device_string.lower()):
                device_name = name
                connect_fn = fn
                break
    
    if connect_fn is None:
        parser.error("Device {!r} not found.".format(args.device))
    
    logging.info("Connecting to device %s.", device_name)
    args.device = connect_fn()
    logging.info("Connected!")


def parse_visibility_arguments(parser, args):
    # Check that filters don't conflict
    if args.visible_object_matchers is not None and args.all:
        parser.error("""
            --all cannot be used with --layer, --id or --class
        """)
    
    # Add default filters
    if not args.all and args.visible_object_matchers is None:
        # Check to see if at least one layer is found and matched by the
        # default regexes before adding that as a filter (if not don't add any
        # filters)
        if args.plot_mode == PlotMode.cut:
            matcher = make_layer_matcher(CUT_LAYER_NAMES_REGEX.pattern)
        else:
            matcher = make_layer_matcher(PLOT_LAYER_NAMES_REGEX.pattern)
        if list(filter(matcher, args.svg.iter())):
            args.visible_object_matchers = [matcher]


def absolute_or_percentage_within(string, low, high):
    """
    Parse a string giving a value as either an absolute float value or a
    percentage. Throws a ValueError when outside the stated range.
    """
    string = string.strip()
    
    # Parse value
    if string.endswith("%"):
        perc = float(string.rstrip("%")) / 100.0
        value = low + ((high - low) * perc)
    else:
        value = float(string)
    
    # Check range
    if not (low <= value <= high):
        raise ValueError("Value must be between {} and {}".format(
            low, high))
    
    return value


def parse_speed_and_force(parser, args):
    try:
        args.speed = absolute_or_percentage_within(
            args.speed,
            args.device.params.tool_speed_min,
            args.device.params.tool_speed_max)
    except ValueError as e:
        parser.error(
            "--speed arguments must be valid ({})".format(str(e)))
    try:
        args.force = absolute_or_percentage_within(
            args.force,
            args.device.params.tool_force_min,
            args.device.params.tool_force_max)
    except ValueError as e:
        parser.error(
            "--force arguments must be valid ({})".format(str(e)))


def parse_regmarks(parser, args):
    if args.regmarks is False:
        # Regmark use is disabled
        args.regmarks = None
    elif isinstance(args.regmarks, list):
        # Manual regmarks
        try:
            svg_width, svg_height = get_svg_page_size(args.svg)
            x, y, width, height = (
                css_dimension_to_mm(
                    d, width_mm=svg_width, height_mm=svg_height)
                for d in args.regmarks
            )
            args.regmarks = RegmarkSpecification(x, y, width, height)
        except ValueError as e:
            parser.error(
                "--manual-regmarks arguments must be valid CSS dimensions ({})".format(
                    str(e)))
    else:
        # Automatic regmarks
        discovered_regmarks = find_regmarks(
            svg_to_outlines(args.svg),
            required_box_size=5.0,
        )
        
        if args.regmarks is True:
            # Automatic regmarks explicitly requested
            if discovered_regmarks is None:
                parser.error("""
                    --regmarks specified but no registration marks were found in
                    the input.  Check that: 1) all regmarks are visible 2) drawn
                    opaque in black 3) drawn with opaque black strokes (including
                    the box) 4) all brackets are the same length and thickness.
                """)
            else:
                args.regmarks = discovered_regmarks
        else:
            # Default mode: use regmarks if discovered but otherwise don't
            # bother
            args.regmarks = discovered_regmarks


def parse_arguments(arg_strings=None):
    """
    Parse a set of commandline arguments returning the resulting options in an
    argparse namespace with the following entries:
    
    * svg: The specified SVG file loaded as an ElementTree.
    * device: The connected :py:class:`py_silhouette.Device`.
    * plot_mode: One of :py:class:`PlotMode`
    * speed: A device speed in mm/sec
    * force: A device force in grams
    * inside_first: bool
    * fast_order: bool
    * regmarks: :py:class:`RegmarkSpecification` or None.
    * include_regmarks: bool
    * colour: list of (r, g, b, a) tuples of colours to be plotted or None if
      no colour filtering is to be applied.
    * visible_object_matchers: None or a list of ElementTree node test
      functions specifying which SVG elements should be made visible.
    """
    parser = make_argument_parser()
    args = parser.parse_args(arg_strings)
    
    if args.list_devices:
        print_device_list(args.use_dummy_device)
        return None
    
    parse_svg_argument(parser, args)
    
    # Guess default plot mode (if not specified)
    if args.plot_mode is None:
        args.plot_mode = guess_plot_mode(args.svg) or PlotMode.cut
        logging.info("Guessed plotting mode as %s", args.plot_mode.value)
    
    # Parse/validate filter arguments
    parse_visibility_arguments(parser, args)
    
    # Select device
    parse_device_arguments(parser, args)
    
    # Parse speed and force settings
    parse_speed_and_force(parser, args)
    
    # Set default inside-first
    if args.inside_first is None:
        args.inside_first = True if args.plot_mode == PlotMode.cut else False
    
    # Set default over-cut distance
    if args.over_cut is None:
        if args.plot_mode is PlotMode.cut:
            args.over_cut = DEFAULT_OVER_CUT 
        else:
            args.over_cut = 0
    
    # Parse and detect regmarks
    parse_regmarks(parser, args)
    
    return args


def args_to_outlines(args):
    """
    Convert the SVG loaded by :py:func:`parse_arguments` into a series of line
    segments according to the command-line options given.
    """
    if args.visible_object_matchers:
        logging.info("Filtering SVG...")
        make_nodes_visible(
            args.svg,
            lambda n: any(m(n) for m in args.visible_object_matchers),
        )
    
    logging.info("Converting SVG into line segments...")
    filtered_lines = [
        line
        for colour, thickness, line in svg_to_outlines(args.svg)
        if (
            # Filter by colour
            (args.colour is None or colour in args.colour) and
            # Remove regmarks
            (
                args.include_regmarks or
                args.regmarks is None or
                not args.regmarks.is_line_part_of_regmark(
                    colour, thickness, line,
                )
            )
        )
    ]
    
    # Offset line coordinates according to registration marks
    if args.regmarks is not None:
        filtered_lines = [
            [
                (
                    x - args.regmarks.x,
                    y - args.regmarks.y,
                )
                for x, y in line
            ]
            for line in filtered_lines
        ]
    
    # Cut inside lines first if reqired
    if args.inside_first:
        logging.info("Ordering inner lines first...")
        grouped_lines = group_inside_first(filtered_lines)
    else:
        grouped_lines = [filtered_lines]
    
    # Optimise cutting order
    optimised_lines = []
    if args.fast_order:
        logging.info("Optimising line order...")
        cur_pos = (0, 0)
        for line_group in grouped_lines:
            optimised_lines.extend(optimise_lines(line_group, cur_pos))
            if optimised_lines:
                cur_pos = optimised_lines[-1][-1]
    else:
        for line_group in grouped_lines:
            optimised_lines.extend(line_group)
    
    # Over-cut closed shapes if required
    if args.over_cut:
        logging.info("Adding over-cut to closed lines...")
        optimised_lines = over_cut_lines(optimised_lines, args.over_cut)
    
    logging.info(
        "Converted SVG into %d lines and %d line segments",
        len(optimised_lines),
        sum(map(len, optimised_lines)) - len(optimised_lines),
    )
    
    return optimised_lines


def zero_on_regmarks(device, regmarks):
    """
    Attempt to find and zero on the registration marks specified. Will allow
    the user to keep re-trying interactively.
    """
    while True:
        logging.info("Searching for registration marks...")
        try:
            device.zero_on_registration_mark(
                regmarks.width,
                regmarks.height,
                box_size=regmarks.box_size,
                line_thickness=regmarks.line_thickness,
                line_length=regmarks.line_length,
            )
            return
        except py_silhouette.RegistrationMarkNotFoundError:
            sys.stderr.write("Registration marks not found, try again?\n")
            sys.stderr.write("Y/n> ")
            char = input().strip().lower()
            if not char or char.startswith("y"):
                continue
            else:
                sys.exit(1)


def main(args=None):
    args = parse_arguments(args)
    
    if args is None:
        return 0
    
    lines = args_to_outlines(args)
    
    if args.regmarks:
        zero_on_regmarks(args.device, args.regmarks)
    
    args.device.set_force(args.force)
    args.device.set_speed(args.speed)
    args.device.set_tool_diameter(
        args.device.params.tool_diameters["Knife"]
        if args.plot_mode == PlotMode.cut else
        args.device.params.tool_diameters["Pen"]
    )
    args.device.flush()
    
    for line in lines:
        for i, (x, y) in enumerate(line):
            pen_down = i > 0
            args.device.move_to(x, y, pen_down)
        args.device.flush()
    
    args.device.move_home()
    args.device.flush()
    
    while args.device.get_state() == py_silhouette.DeviceState.moving:
        time.sleep(0.5)
    
    return 0


if __name__ == "__main__":
    main()
